#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2003 Free Software Foundation
#
# FILE:
# adodbapi/Driver.py
#
# DESCRIPTION:
# Driver to provide access to data via adodbapi Driver
#
# NOTES:
# Only works under Win32... requires the win32all extensions.
# (http://aspn.activestate.com/ASPN/Downloads/ActivePython/Extensions/Win32all)
#
# Home page: http://sourceforge.net/projects/adodbapi
#
#   Supported attributes (via connections.conf or <database> tag)
#
#     oledb_provider   (required)
#     data_source      (required)
#     initial_catalog   (optional for SQL Server)
#     network_library (optional for SQL Server)
#     data_provider    (optional for SQL Server)
#
# You can find more connection strings here:
# http://www.able-consulting.com/MDAC/ADO/Connection/OLEDB_Providers.htm
#

_examples = """
[access]
comment = MS Access database
provider = adodbapi
oledb_provider = Microsoft.Jet.OLEDB.4.0
data_source = C:\mydb.mdb

[sqlserver]
comment = MS SQL Server database
provider = adodbapi
oledb_provider = sqloledb
data_source = myServerName
initial_catalog = myDatabaseName
"""

from gnue.common.datasources import GDataObjects, GConditions, GConnections
from gnue.common.apps import GDebug
from gnue.common.datasources.drivers.DBSIG2.Driver \
   import DBSIG_RecordSet, DBSIG_ResultSet, DBSIG_DataObject, \
          DBSIG_DataObject_SQL, DBSIG_DataObject_Object

try:
  import adodbapi as SIG2api
except ImportError, message:
  tmsg = _("Driver not installed: adodbapi driver\n\n[%s") % message
  raise GConnections.AdapterNotInstalled, tmsg

import win32com

DESCRIPTION_NAME            = 0
DESCRIPTION_TYPE_CODE       = 1
DESCRIPTION_DISPLAY_SIZE    = 2
DESCRIPTION_INTERNAL_SIZE   = 3
DESCRIPTION_PRECISION       = 4
DESCRIPTION_SCALE           = 5
DESCRIPTION_NULL_OK         = 6

TABLE_CATALOG = 0
TABLE_SCHEMA = 1 
TABLE_NAME = 2
TABLE_TYPE = 3

adSchemaTables = 20


class Adodbapi_RecordSet(DBSIG_RecordSet):
  pass


class Adodbapi_ResultSet(DBSIG_ResultSet):
  def __init__(self, dataObject, cursor=None, defaultValues={}, masterRecordSet=None):
    DBSIG_ResultSet.__init__(self, dataObject, \
            cursor, defaultValues, masterRecordSet)
    self._recordSetClass = Adodbapi_RecordSet


class Adodbapi_DataObject(DBSIG_DataObject):
  def __init__(self):
    DBSIG_DataObject.__init__(self)
    self._DatabaseError = SIG2api.DatabaseError
    self._resultSetClass = Adodbapi_ResultSet

  def connect(self, connectData={}):
    GDebug.printMesg(1,"Adodbapi database driver initializing")

    try:
      connPar = {'Provider': connectData['oledb_provider'],
                      'Data Source' : connectData['data_source'],
                      'User Id' : connectData['_username'],
                      'Password' : connectData['_password']
                     }
      # optional parameters for MS SQL Server
      for gnueName, oledbName in [('initial_catalog', 'Initial Catalog'),
                                              ('network_library', 'Network Library'),
                                              ('data_provider', 'Data Provider'),
                                              ]:
          if connectData.has_key(gnueName):
              connPar[oledbName] = connectData[gnueName]

      connStr = ';'.join([r"%s=%s" % (k,v) for k,v in connPar.items()])
      GDebug.printMesg(1,'Adodbapi connection string="%s"' % connStr)

      self._dataConnection = SIG2api.connect(connStr)

    except self._DatabaseError, value:
      raise GDataObjects.LoginError, value

    self._postConnect()
    

  def _postConnect(self):
    self.triggerExtensions = TriggerExtensions(self._dataConnection)


  def _createResultSet(self, conditions={}, readOnly=0, masterRecordSet=None,sql=""):

    # Used by drivers with a unique id (like rowid)
    if not self._primaryIdChecked: self._checkForPrimaryId()

    try:
      cursor = self._dataConnection.cursor()

      cursor.arraysize = self.cache
      cursor.execute(self._buildQuery(conditions, additionalSQL=sql))

      # pull a record count 
      if self._strictQueryCount:
#        recordCount = cursor.rowcount
#        #disable the count query and see if anyone screams
#        #recordCount = self._getQueryCount(conditions,sql)

        #adodbapi screams :(
        recordCount = self._getQueryCount(conditions,sql)
        
    except self._DatabaseError, err:
      raise GDataObjects.ConnectionError, err

    rs = self._resultSetClass(self, cursor=cursor, masterRecordSet=masterRecordSet)
    if self._strictQueryCount:
      rs._recordCount = recordCount
    if readOnly:
      rs._readonly = readOnly

    return rs


  #
  # Schema (metadata) functions
  #

  # TODO: See postgresql for an example of what these functions do.

  # Return a list of the types of Schema objects this driver provides
  def getSchemaTypes(self):
    return [('view',_('Views'),1),
            ('table',_('Tables'),1)]

  # Return a list of Schema objects
  def getSchemaList(self, type=None):
    QueryType = adSchemaTables
    rs = self._dataConnection.adoConn.OpenSchema(QueryType)

    rs.MoveFirst()
    tables = []
    while not rs.EOF:
      if str(rs.Fields(TABLE_TYPE)) == 'TABLE':
        tables.append((str(rs.Fields(TABLE_NAME)),'table'))
      elif str(rs.Fields(TABLE_TYPE)) == 'VIEW':
        tables.append((str(rs.Fields(TABLE_NAME)),'view'))
      rs.MoveNext()
      
    list = []
    for t in tables:
      list.append(GDataObjects.Schema(attrs={'id':t[0], 'name':t[0],
                         'type':t[1],
			 'primarykey': self.__getPrimaryKey(t[0])},
                         getChildSchema=self.__getFieldSchema))
    return list

  # Find a schema object with specified name
  def getSchemaByName(self, name, type=None):
    
    schema = GDataObjects.Schema(attrs={'id':name, 'name':name,
                         'type':'table',
			 'primarykey': self.__getPrimaryKey(name)},
                         getChildSchema=self.__getFieldSchema)
    return schema

  # Return a list of fields (for _buildDeleteStatement and for _buildUpdateStatement)
  def __getPrimaryKey(self, relname):
    # TODO
    return []

  # Get fields for a table
  def __getFieldSchema(self, parent):
    statement = "select * from %s where 0=1" % (parent.name)

    cursor = self._dataConnection.cursor()
    cursor.execute(statement)

    list = []

    for d in cursor.description:
      try:
        nativetype = d[DESCRIPTION_TYPE_CODE]
      except AttributeError:
        nativetype='unknown'
      
      attrs={'id':d[DESCRIPTION_NAME],
                 'name':d[DESCRIPTION_NAME],
                 'type':'field',
                 'nativetype': nativetype,
                 'required': d[DESCRIPTION_NULL_OK]==0,
                 'length': d[DESCRIPTION_INTERNAL_SIZE]}

      if nativetype == SIG2api.NUMBER:
        attrs['datatype']='number'
        attrs['precision']=d[DESCRIPTION_SCALE]
        attrs['length']=d[DESCRIPTION_PRECISION]
      elif nativetype == SIG2api.DATETIME:
        attrs['datatype']='date'
      else:
        attrs['datatype']='text'

      list.append(GDataObjects.Schema(attrs=attrs))

    cursor.close()
    return list


class Adodbapi_DataObject_Object(Adodbapi_DataObject, \
      DBSIG_DataObject_Object):

  def __init__(self):
    Adodbapi_DataObject.__init__(self)

  def _buildQuery(self, conditions={},forDetail=None,additionalSQL=""):
    return DBSIG_DataObject_Object._buildQuery(self, conditions,forDetail,\
                                                additionalSQL)


class Adodbapi_DataObject_SQL(Adodbapi_DataObject, \
      DBSIG_DataObject_SQL):
  def __init__(self):
    # Call DBSIG init first because Adodbapi_DataObject needs to overwrite
    # some of its values
    DBSIG_DataObject_SQL.__init__(self)
    Adodbapi_DataObject.__init__(self)

  def _buildQuery(self, conditions={}):
    return DBSIG_DataObject_SQL._buildQuery(self, conditions, forDetail,\
                                            additionalSQL)


#
#  Extensions to Trigger Namespaces
#
class TriggerExtensions:

  def __init__(self, connection):
    self.__connection = connection



######################################
#
#  The following hashes describe
#  this driver's characteristings.
#
######################################

#
#  All datasource "types" and corresponding DataObject class
#
supportedDataObjects = {
  'object': Adodbapi_DataObject_Object,
  'sql':    Adodbapi_DataObject_SQL
}
