#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001-2003 Free Software Foundation
#
# FILE:
# GComm.py
#
# DESCRIPTION:
# Class that provides a basis for GNUe Comm clients & servers.
#
# NOTES:
#

import string
from gnue.common.apps import GDebug
from gnue.common.utils.FileUtils import dyn_import, openResource


#####################################################################
#
# Helper functions to load appropriate driver
# 
# TODO: add to FileUtils as they are almost the same function as in GConnection.py

#
# Load the correct driver from gnue/common/datasources/drivers/*/
#

from gnue.common.rpc.drivers import DRIVERS as ALLDRIVERS

###################################################
# 
#  load driver from path which passes test 'test'
#  f.e. _get_driver('xmlrpc','gnue.common.rpc.drivers',lambda x: return x.CLIENT==1))

def _get_driver(drivername, baselist, path, test=lambda x:1):

  d = drivername.split('.')
  basedriver = d[0]
  if len(d) > 1:
    extradriver = "." + string.join(d[1:],'.')
  else:
    extradriver = ""

  driver = None

  basemodule = _find_base_driver(basedriver, baselist, [path])
  GDebug.printMesg(1,'Using %s as base driver for %s' %  (basemodule, driver))

  if basemodule:
    driver = _import_driver(basemodule + extradriver, test)

  if not driver:
    tmsg = _("No rpc driver found for provider type '%s'") % drivername
    raise Exception, tmsg

  return driver

def _find_base_driver (driver, modules, path=[]):
  if driver in modules:
    return string.join(path + [driver],'.')
  else:
    for module in modules:
      try:
        m = dyn_import (string.join(path + [module],'.')).DRIVERS
        rs = _find_base_driver(driver, m, path + [module])
        if rs:
          return rs
      except (AttributeError, ImportError), err:
        pass


def _import_driver (drivername,test):
  driver = None
  try:
    driver = dyn_import("%s.Driver" % (drivername))
  except:
    GDebug.printMesg(1,'%s is not a dbdriver' % ( drivername))
    try:
      drivers = dyn_import("%s" % (drivername)).DRIVERS
      for d in drivers:
        driver = _import_driver(drivername + "." + d, test )
        try:
          if driver:
            if test(driver):
              return driver
        except Exception,msg:
          print msg
    except (ImportError, AttributeError):
      GDebug.printMesg(1,'%s does not contain drivers' % (drivername))
  return driver




##############################################################################
#
# Attach to a client driver
#
def attach(interface, params):
  try:
    driver = _get_driver(interface, ALLDRIVERS, 'gnue.common.rpc.drivers',lambda x:x.CLIENT==1 )
    if driver == None:
      raise ImportError
  except ImportError, mesg:
    GDebug.printMesg(1, \
       "Unable to import GComm Adapter '%s': \n%s" % (interface,mesg))
    raise InvalidAdapter, mesg

  adapter = driver.ClientAdapter(params)

  return adapter



##############################################################################
#
# Bind/export our services
#
# Attributes:
#
# rpcdef     The RPC definition loaded from an XML file. This can be
#              1) a string that contains the location of the XML file
#                 (can be a URL or pathname)
#              2) a file-handle (or file-like handle; e.g. StringBuffer)
#                 referencing an XML definition
#              3) a GnuRpc object created from GComm.loadDefinition()
#
# drivers    A dictionary that defines all the transports to be used. The
#            dictionary is of the form: {driver:params} where:
#              1) driver: A string containing the name of the commdriver
#                 (e.g., 'xmlrpc' or 'corba')
#              2) params: A dictionary containing parameters specific to
#                 the commdriver (e.g., {'port':'8888'} )
#
# bindings   A dictionary containing binding definitions. The dictionary
#            is of the form: {server:handler} where:
#              1) service: Name of the service being bound/exposed
#              2) handler: Method that when called returns a class 
#                 instance i.e. {'DonutPlace',getDonutPlaceObj}
#
def bind(rpcdef, drivers, bindings):

  if type(rpcdef) == type(""):

    # Application supplied the location of their rpcdef.. load it
    fin = openResource(rpcdef)
    mapping = loadDefinition(fin,1)
    fin.close()

  elif rpcdef.hasattr("read"):

    # Application provided us a file-like object
    mapping = loadDefinition(rpcdef)

  else:

    # Otherwise, they must have specified a GnuRpc object
    mapping = rpcdef


  servers = {}

  for interface in drivers.keys():
    params = drivers[interface]

    try:
      driver = _get_driver(interface, ALLDRIVERS, 'gnue.common.rpc.drivers',lambda x:x.SERVER==1 )
      if driver == None:
        raise ImportError
    except ImportError, mesg:
      GDebug.printMesg(1, \
                       "Unable to import GComm Adapter '%s': \n%s" % (interface,mesg))
      raise InvalidAdapter, mesg

    adapter = driver.ServerAdapter(mapping, bindings, params)
    servers[interface] = adapter

  return servers



#
# To keep everything that a client/server app will need in this
# file, any convenience functions from the commdrivers directory.
# (i.e., a user app should never need to access the commdrivers
# directory.)
#

#
# Load an XML-based interface definition
#
from gnue.common.rpc.parser.Parser import loadDefinition




##############################################################################
#
# Exception hierarchy:
#
# Error
# |__ ProgrammingError
# |   |__ InvalidService
# |__ CommunicationsError
# |__ UserError (base for user-defined errors)
#

#
# The basis of all GComm exceptions
#
class Error(StandardError):
  pass

#
# The requested adapter does not exist...
#
class InvalidAdapter(Error):
  pass

#
# The requested adapter could not initialize. Perhaps the
# supplied parameters do not point to a valid server, etc.
#
class AdapterInitializationError(Error):
  pass

#
# The parameters supplied to the adapter are not in the
# correct format, or all the needed parameters were not
# supplied.
#
class AdapterConfigurationError(AdapterInitializationError):
  pass

#
# The application requested a "client" adapter, but the
# adapter only supports "server" adapters.
#
class NoClientAdapter(Error):
  pass

#
# The application requested a "server" adapter, but the
# adapter only supports "client" adapters.
#
class NoServerAdapter(Error):
  pass

#
# TBD..
#
class ProgrammingError(Error):
  pass

#
# A client application requested a service that is not
# exposed by the server.
#
class InvalidService(ProgrammingError):
  pass

#
# A communications error interfered with the client and
# server's transaction.
#
class CommunicationsError(Error):
  pass


#
# An error which has occured on the server and which is
# raised again on the server
#
class DistantError(Error):
  pass


#
# Hmm... I had something in mind when I put this down...
#
class UserError(Error):
  pass



#
# Some conversion functions to allow a client to describe the type of
# the argument that should be passed instead of autodetection
# TODO: should be moved to GTypeCast
#
# i.e. Donutplace.send(base64(photo.jpg))

class base64:
  def __init__(self,value):
    self._type='base64'
    self._value=value

class binary:
  def __init__(self,value):
    self._type='binary'
    self._value=value
    
class date:
  def __init__(self,value):
    self._type='date'
    self._value=value

# value has to be a 6 tuple: ex: datetime(1,1,1,1,1,1)
class datetime:
  def __init__(self,value):
    self._type='date'
    self._value=value    

class boolean:
  def __init__(self,value):
    self._type='boolean'
    self._value=value

class integer:
  def __init__(self,value):
    self._type='integer'
    self._value=value

class number:
  def __init__(self,value):
    self._type='number'
    self._value=value

class string_type:
  def __init__(self,value):
    self._type='string'
    self._value=value

class none:
  def __init__(self,value):
    self._type='none'
    self._value=value
