#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2002-2003 Free Software Foundation
#
# FILE:
# Scripter.py
#
# DESCRIPTION:
# Schema definition scripter
#
# NOTES:
# Beware the Scriptor() class -- it smells like a billy goat.
#

from gnue.common import VERSION
from gnue.common.schema import GSParser
from gnue.common.utils.FileUtils import openResource, dyn_import
from gnue.common.apps.GClientApp import GClientApp
from processors import vendors

import sys
import os

class ScripterRunner(GClientApp):
  #
  # GClientApp() overrides
  #
  VERSION = VERSION
  COMMAND = "gnue-schema"
  NAME = "GNUe Schema Scripter"
  USAGE = "[options] file [old-schema]"
  COMMAND_OPTIONS = [
      [ 'drop_tables',None,'drop-tables', 0, None, None,
          'Generate commands to drop relevant tables.  * NOT IMPLEMENTED'],
      [ 'ignore_schema','S','no-create', 0, None, None,
          'Do not generate schema creation code.  * NOT IMPLEMENTED'],
      [ 'ignore_data','D','no-data', 0, None, None,
          'Do not generate data insertion code.  * NOT IMPLEMENTED'],
      [ 'upgrade_schema','u','upgrade-schema', 0, None, None,
          'Generate code to upgrade an older version of a schema to '
          'the recent version. You must specify a previous schema with on the '
          'command line.  * NOT IMPLEMENTED'],
      [ 'upgrade_data','U','upgrade-data', 0, None, None,
          'Generate code to upgrade an older version of schema data to '
          'the recent version. You must specify a previous schema with on the '
          'command line.  * NOT IMPLEMENTED'],
      [ 'list_vendors','l','list-vendors', 0, None, None,
          'List all supported vendors.'],
      [ 'output','o','output', 1, None, 'dest',
          'The destination for the created schemas. This can be in several '
          'formats. If <dest> is a file name, then output is written to this '
          'file. If <dest> is a directory, then <dest>/<Vendor>.sql is created. '
          'The default is to create <Vendor>.sql in the current directory. '
          'NOTE: the first form (<dest> as a filename) is not supported for '
          '--vendors all.' ],
      [ 'vendor','v','vendor', 1, 'all', 'vendor',
          'The vendor to create a script for. If <vendor> is "all", then '
          'scripts for all supported vendors will be created. <vendor> can '
          'also be a comma-separated list.'],
      ]
  SUMMARY = \
     "GNUe Schema Scripter creates SQL files based on GNUe Schema Definitions."

  def __init__(self, connections=None):
    ConfigOptions = {}
    GClientApp.__init__(self, connections,'schema',ConfigOptions)
  
  #
  # Run
  #
  def run(self):

    globals().update(self.OPTIONS)

    # List vendors, if requested
    if list_vendors:
      self.listVendors()
      sys.exit()

    upgrading = 0

    # Do some sanity checks on the options
    if upgrade_schema or upgrade_data:
      upgrading = 1

      if drop_tables:
        self.handleStartupError('--drop-tables is not compatable with the '
               '--upgrade-?? options. Unable to create an upgrade schema.')
      try:
        old_schema = self.ARGUMENTS[1]
      except IndexError:
        self.handleStartupError('An --upgrade-?? option was selected, but '
               'only one source file was specified. Unable to create '
               'an upgrade schema.')

    elif drop_tables and ignore_schema:
      # If we are dropping tables and ignoring schema,
      # then we can't possibly want the data.
      ignore_data = 1


    if vendor.lower() == 'all':
      vens = vendors

    else:
      vens = vendor.split(',')

    if len(vens) > 1 and (output and not os.path.isdir(output)):
      self.handleStartupError('--output cannot reference a file '
                              'if multiple vendors are specified.')

    # Assign input file
    try:
      schema = self.ARGUMENTS[0]
      input = openResource(schema)
    except IndexError:
      self.handleStartupError('No source file was specified.')
    except IOError:
      self.handleStartupError('Unable to open requested file: %s' % schema)

    print
    scripter = Scripter(schema)
    for ven in vens:
      if not output:
        outfile = self.getVendorName(ven) + '.sql'
      elif os.path.isdir(output):
        outfile = os.path.join(output, self.getVendorName(ven) + '.sql')
      else:
        outfile = output

      try:
        out = open(outfile,'w')
      except IOError, mesg:
        self.handleStartupError('Unable to open destination file: %s\n\n%s' % (outfile,mesg))

      print "Writing schema to %s ..." % outfile

      try:
        scripter.writeSql(ven, out)
      except:
        print "WARNING: Unable to create a schema for %s" % ven
        raise

      out.close()

  #
  # Get a vendor name
  #
  def getVendorName(self, vendor):
    return dyn_import('gnue.common.schema.scripter.processors.%s' % vendor).name

  #
  # List Vendors
  #
  def listVendors(self):
    header = "%s\nVersion %s" % (self.NAME, self.VERSION)
    print
    print header
    print
    print "Supported Database Vendors"
    print "--------------------------"

    modules = {}
    maxsize = 0
    for vendor in vendors:
      maxsize = max(maxsize, len(vendor))
      try:
        modules[vendor] = dyn_import('gnue.common.schema.scripter.processors.%s' % vendor)
      except ImportError:
        pass

    srt = modules.keys()
    srt.sort()
    for vendor in srt:
      print vendor.ljust(maxsize+4), modules[vendor].description

    print






############################################################
#
# The actual worker class
#
class Scripter:
  def __init__(self, source, oldsource=None):

    if hasattr(source,'readline'):
      input = source
    else:
      input = openResource(source)

    self.schema = GSParser.loadFile(input)


  def writeSql(self, format, destination=sys.stdout):
    self.processor = dyn_import('gnue.common.schema.scripter.processors.%s' % format).Processor()
    self.destination = destination

    self.schema.walk(self._walkTables)

  def _walkTables(self, object):
    if object._type == 'GSTable':
      self.processor.comment (self.destination,"\nCreate '%s' table\n" % object.name)
      self.destination.write('\n')
      self._fields = []
      self._postfields = []
      self._pretable = []
      self._posttable = []
      self._pkfields = []
      self._pkname = ""
      object.walk(self._walkTable, tablename=object.name)
      if len(self._pkfields):
        self._handlePK(object.name)
      self.destination.write(self.processor.createTable(object.name, self._fields + self._postfields, self._pretable, self._posttable))
      self.destination.write('\n\n')

  def _walkTable(self, object, tablename):
    if object._type == 'GSField':
      exec "self._handlePrePost(self.processor.createField(object.name,tablename,self.processor.%s(object), object))" % object.type in locals()
    elif object._type == 'GSPKField':
      self._pkfields.append(object.name)
    elif object._type == 'GSIndex':
      self._indexfields = []
      object.walk(self._walkIndex, tablename)
      self._handlePrePost(self.processor.createIndex(object.name,tablename,self._indexfields, object.unique))


  def _walkIndex(self, object, tablename):
    if object._type == 'GSIndexField':
      self._indexfields.append(object.name)


  def _handlePK(self, tablename):
    self._handlePrePost(self.processor.createPrimaryKey(self._pkname,tablename, self._pkfields))


  # Butt-ugly stuff
  def _handlePrePost(self, results):
    field, extra, pretable, posttable = results
    if field:
      self._fields.append(field)
    if extra:
      self._postfields += extra
    if pretable:
      self._pretable += pretable
    if posttable:
      self._posttable += posttable


if __name__ == '__main__':
  ScripterRunner().run()
